import pandas as pd


def load_covid():
    covid = pd.read_csv("https://raw.githubusercontent.com/owid/covid-19-data/master/public/data/owid-covid-data.csv")

    target = 'new_cases'
    date = 'date'

    dataset = covid[(covid['location'] == 'World')].copy()[[target, date]]
    dataset[date] = pd.to_datetime(dataset[date])
    dataset.index = dataset[date]

    dataset['month'] = dataset['date'].dt.month
    dataset = dataset.drop(columns=['date'])

    return {
        'target': target,
        'dataset': dataset,
    }



import unittest

import pandas as pd
from matplotlib import pyplot as plt

from sklearn_ts.features.seasonality import generate_fourier, add_fourier_to_X


class FourierTestCase(unittest.TestCase):
    def test_Fourier_simple(self):
        fourier = generate_fourier(0, 2, periods=[7], N=[1], with_intercept=False)
        fourier_ref = pd.DataFrame({
            'fourier_sin_7_1': [0.0, 0.78183],
            'fourier_cos_7_1': [1.0, 0.62349],
        })
        self.assertIsNone(pd.testing.assert_frame_equal(fourier, fourier_ref, check_less_precise=True))

    def test_Fourier_add_toX(self):
        X = pd.DataFrame({'trend': [0, 1]}, index=['2021-07-01', '2021-07-02'])
        X_with_fourier = add_fourier_to_X(X, periods=[7], N=[1], with_intercept=False)
        X_with_fourier_ref = pd.DataFrame({
            'trend': [0, 1],
            'fourier_sin_7_1': [0.0, 0.78183],
            'fourier_cos_7_1': [1.0, 0.62349],
        }, index=['2021-07-01', '2021-07-02'])

        self.assertIsNone(pd.testing.assert_frame_equal(X_with_fourier, X_with_fourier_ref,
                                                        check_less_precise=True))

    def test_Fourier_plotting(self):
        fourier = generate_fourier(0, 24, periods=[7, 12], N=[1, 1], with_intercept=False)

        fig, axes = plt.subplots(nrows=1, ncols=1, figsize=(10, 6))
        fourier.plot.line(ax=axes,
                          color=['orange', 'orange', 'blue', 'blue'],
                          style=['-', '--', '-', '--'],
                          title='Fourier components'
                          )
        fig.savefig(f'Fourier.png')

        self.assertEqual(fourier.shape[1], 4)



import unittest

from tests.mape_test import MAPETestCase

from tests.validator_test import ValidatorTestCase


def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(MAPETestCase())
    test_suite.addTest(ValidatorTestCase('test_widget_resize'))
    return test_suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



import unittest

from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import MinMaxScaler
from sklearn.svm import SVR

from sklearn_ts.datasets.covid import load_covid
from sklearn_ts.validator import check_model


class ValidatorTestCase(unittest.TestCase):

    def test_regression(self):
        dataset = load_covid()['dataset']

        params = {'fit_intercept': [False]}
        regressor = LinearRegression(fit_intercept=False)

        results = check_model(
            regressor, params, dataset,
            target='new_cases', features=['month'], categorical_features=[], user_transformers=[],
            h=30, n_splits=5, gap=30,
            plotting=True
        )

        self.assertEqual(len(results.keys()), 8)

    def test_SVR(self):
        dataset = load_covid()['dataset']

        params = {'C': [1.0]}
        regressor = SVR()

        results = check_model(
            regressor, params, dataset,
            target='new_cases', features=['month'], categorical_features=[], user_transformers=[('mm', MinMaxScaler())],
            h=14, n_splits=2, gap=14,
            plotting=True
        )

        self.assertEqual(len(results.keys()), 8)



from matplotlib import pyplot as plt


def pretty_plot(chart, save=False):
    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(10, 6))

    chart(ax)
    ax.get_legend().remove()
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)

    if save:
        fig = ax.get_figure()
        fig.savefig(f'{chart.__name__}.png')



from sklearn.base import BaseEstimator, RegressorMixin
from sktime.forecasting.naive import NaiveForecaster
import numpy as np
import pandas as pd


class NaiveRegressor(BaseEstimator, RegressorMixin):

    def __init__(self, strategy='last', sp=7):
        self.strategy = strategy
        self.sp = sp

        self.model = None
        self.y_ = None

    def fit(self, X, y):
        forecaster = NaiveForecaster(strategy=self.strategy, sp=self.sp)

        # TODO inference frequency
        y = y.asfreq(pd.infer_freq(y.index))
        self.y_ = y
        forecaster.fit(y)
        self.model = forecaster
        return self

    def predict(self, X):
        # TODO not valid for fitting to train
        return self.model.predict(np.arange(X.shape[0]) + 1)

    def get_params(self, deep=True):
        return {'strategy': self.strategy, 'sp': self.sp}

    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            setattr(self, parameter, value)
        return self



import numpy as np
import pandas as pd


def add_fourier_to_X(X, t_start=0, periods=[7], N=[10], with_intercept=False):
    fourier = generate_fourier(t_start, X.shape[0] + t_start, periods, N, with_intercept)
    fourier.index = X.index
    return X.join(fourier)


def generate_fourier(t_start, t_end, periods=[7], N=[10], with_intercept=False):
    t = np.arange(t_start, t_end)

    fourier_dict = {}
    for i, period in enumerate(periods):
        n_vec = np.arange(start=0 if with_intercept else 1, stop=N[i] if with_intercept else N[i] + 1)
        for n in n_vec:
            x = 2 * np.pi * n * t / period
            fourier_dict[f'fourier_sin_{period}_{n}'] = np.sin(x)
            fourier_dict[f'fourier_cos_{period}_{n}'] = np.cos(x)

    return pd.DataFrame(fourier_dict)



import unittest

import pandas as pd

from sklearn_ts.validator import mean_absolute_percentage_error


class MAPETestCase(unittest.TestCase):
    def test_MAPE_equals(self):
        y_true = pd.Series([0, 1])
        y_pred = pd.Series([0, 1])
        mape = mean_absolute_percentage_error(y_true, y_pred)
        self.assertEqual(mape, 0)

    def test_MAPE_zeros(self):
        y_true = pd.Series([0, 0, 1])
        y_pred = pd.Series([0.2, 0.4, 1])
        mape = mean_absolute_percentage_error(y_true, y_pred)
        self.assertEqual(round(mape, 10), 0.2)

    def test_MAPE_zeros_epsilon(self):
        y_true = pd.Series([0, 0, 1])
        y_pred = pd.Series([0.2, 0.4, 1])
        mape = mean_absolute_percentage_error(y_true, y_pred, zeros_strategy='epsilon')
        self.assertGreater(mape, 10**6)



import numpy as np
import pandas as pd


# TODO gap for example each month for half a year
def custom_split(data, h, n_splits, gap):
    """
    Splitting data into time-CV parts
    :param data: assumed to be already ordered
    :param h: forecasting horizon
    :param n_splits: number of splits
    :param gap: gap between performing forecasts, f.e. [n-h, n), [n-2(h+gap), n-h+gap), ...
    :return: splitted datasets, splitted indices (preferably: dates)
    """
    splits = []
    dates = []
    for i in range(0, n_splits):
        train = np.arange(data.shape[0] - h - i * gap)
        test = np.arange((data.shape[0] - h - i * gap), (data.shape[0] - i * gap))

        splits.append((train, test))
        dates.append((data.iloc[train].index, data.iloc[test].index))

    return splits, dates


def separate_data(dataset, target, features, categorical_features):
    y = dataset[target]
    X = dataset[features]
    X_dummies = pd.get_dummies(dataset[features], columns=categorical_features)
    return y, X, X_dummies


def split(dataset, h):
    train = dataset[:(dataset.shape[0]-h)].copy()
    test = dataset[(dataset.shape[0]-h):].copy()
    return train, test



from keras.callbacks import EarlyStopping
from keras.layers import Dense
from keras.layers import LSTM
from keras.models import Sequential
from keras.optimizers import Adam

from tcn import TCN
from tensorflow.keras.layers import Dense
from tensorflow.keras.models import Sequential

from sklearn.base import BaseEstimator, RegressorMixin
import numpy as np


class ANNRegressor(BaseEstimator, RegressorMixin):

    def __init__(self, learning_rate=0.001, loss='mean_squared_error'):
        self.loss = loss
        self.learning_rate = learning_rate

        self.history = None
        self.model = None

    def fit(self, X, y):
        model = Sequential()
        model.add(Dense(12, input_dim=X.shape[1], activation='relu'))
        model.add(Dense(8, activation='relu'))
        model.add(Dense(1, kernel_initializer='normal'))

        adam = Adam(learning_rate=self.learning_rate)
        model.compile(loss='mean_squared_error', optimizer=adam)

        es = EarlyStopping(monitor='val_loss', patience=20)
        self.history = model.fit(X, y, epochs=100, batch_size=10, verbose=0, callbacks=[es], validation_split=0.1)

        self.model = model
        return self

    def predict(self, X):
        return np.squeeze(self.model.predict(X))

    def get_params(self, deep=True):
        return {'loss': self.loss, 'learning_rate': self.learning_rate}

    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            setattr(self, parameter, value)
        return self


# TODO states, memory and join with regular nn
class LSTMRegressor(BaseEstimator, RegressorMixin):
    # TODO automatic feature creation
    # X_pd = pd.DataFrame(X)
    # cols = X_pd.columns
    # lags = [1, 2, 3]
    # lagged_features = []
    # for lag in lags:
    #   for feature in cols:
    #     X_pd[f'{feature}_{lag}_lag'] = X_pd[feature].shift(lag)
    #     lagged_features.append(f'{feature}_{lag}_lag')
    # trainX = np.reshape(X_pd[lagged_features].values, (X.shape[0], len(lags), len(cols)))
    # X_pd.head()

    def __init__(self, learning_rate=0.001, loss='mean_squared_error'):
        self.loss = loss
        self.learning_rate = learning_rate

        self.history = None
        self.model = None

    def fit(self, X, y):
        model = Sequential()
        # model.add(LSTM(64, return_sequences=True))
        model.add(LSTM(32, activation='relu'))  # , input_shape=(X.shape[1], 1)
        model.add(Dense(1))

        adam = Adam(learning_rate=self.learning_rate)
        model.compile(loss='mean_squared_error', optimizer=adam)

        # reshape input to be [samples, time steps, features]
        reshaped_X = np.reshape(X, (X.shape[0], X.shape[1], 1))
        reshaped_y = np.reshape(y.values, (y.shape[0], 1))

        es = EarlyStopping(monitor='val_loss', patience=20)
        self.history = model.fit(reshaped_X, reshaped_y,
                                 epochs=400, batch_size=1, verbose=0, callbacks=[es], validation_split=0.1)

        self.model = model
        return self

    def predict(self, X):
        # TODO train and predict several times to have average
        reshaped_X = np.reshape(X, (X.shape[0], X.shape[1], 1))
        return np.squeeze(self.model.predict(reshaped_X))

    def get_params(self, deep=True):
        return {'loss': self.loss, 'learning_rate': self.learning_rate}

    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            setattr(self, parameter, value)
        return self


# https://github.com/philipperemy/keras-tcn
class TCNRegressor(BaseEstimator, RegressorMixin):

    def __init__(self, learning_rate=0.001, loss='mean_squared_error'):
        self.loss = loss
        self.learning_rate = learning_rate

        self.history = None
        self.model = None

    def fit(self, X, y):
        tcn_layer = TCN(input_shape=(X.shape[1], 1))

        m = Sequential([
            tcn_layer,
            Dense(1)
        ])

        adam = Adam(learning_rate=self.learning_rate)
        m.compile(optimizer=adam, loss='mse')

        reshaped_X = np.reshape(X, (X.shape[0], X.shape[1], 1))
        reshaped_y = np.reshape(y.values, (y.shape[0], 1))

        es = EarlyStopping(monitor='val_loss', patience=20)
        self.history = m.fit(reshaped_X, reshaped_y, epochs=500, validation_split=0.1, verbose=0, callbacks=[es])

        self.model = m
        return self

    def predict(self, X):
        reshaped_X = np.reshape(X, (X.shape[0], X.shape[1], 1))
        return np.squeeze(self.model.predict(reshaped_X))

    def get_params(self, deep=True):
        return {'loss': self.loss, 'learning_rate': self.learning_rate}

    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            setattr(self, parameter, value)
        return self



from setuptools import setup
from setuptools import find_packages


setup(
    name='sklearn-ts',
    packages=find_packages(),
    version='0.0.4',
    description='Package for time series forecasting',
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    keywords="time series forecasting deep learning machine learning regression ARIMA ETS",
    url='https://github.com/lady-pandas/sklearn-ts',
    author='Marta Markiewicz',
    author_email='m.markiewicz.pl@gmail.com',
    license='MIT',
    zip_safe=False
)



import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin
from statsmodels.tsa.exponential_smoothing.ets import ETSModel


class ExpSmoothingRegressor(BaseEstimator, RegressorMixin):
    # https://www.statsmodels.org/devel/generated/statsmodels.tsa.exponential_smoothing.ets.ETSModel.html

    def __init__(self, error='add', trend=None, damped_trend=False, seasonal=None, seasonal_periods=None):
        self.error = error
        self.trend = trend
        self.damped_trend = damped_trend
        self.seasonal = seasonal
        self.seasonal_periods = seasonal_periods

        self.model = None
        self.predictions= None

    def fit(self, X, y):
        model = ETSModel(y, error=self.error, trend=self.trend, seasonal=self.seasonal,
                         damped_trend=self.damped_trend, seasonal_periods=self.seasonal_periods)
        self.model = model.fit(maxiter=10000)
        return self

    def predict(self, X):
        X = np.ndarray.flatten(X)
        self.predictions = self.model.get_prediction(start=min(X), end=max(X)).summary_frame(alpha=0.05)
        return self.predictions['mean']

    def get_params(self, deep=True):
        return {"error": self.error, 'trend': self.trend, 'damped_trend': self.damped_trend, 'seasonal': self.seasonal,
                'seasonal_periods': self.seasonal_periods}

    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            setattr(self, parameter, value)
        return self



import math

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


from sklearn.compose import ColumnTransformer
from sklearn.metrics import make_scorer, mean_absolute_error, mean_squared_error
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import FunctionTransformer

from sklearn_ts.splitter import split, separate_data, custom_split


def mean_absolute_percentage_error(y_true, y_pred, zeros_strategy='mae'):
    """
    Similar to sklearn https://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_absolute_error.html
    with options for behaviour for around zeros
    :param y_true:
    :param y_pred:
    :param zeros_strategy:
    :return:
    """
    epsilon = np.finfo(np.float64).eps
    if zeros_strategy == 'epsilon':
        ape = np.abs(y_pred - y_true) / np.maximum(np.abs(y_true), epsilon)
    elif zeros_strategy == 'mae':
        ae = np.abs(y_pred - y_true)
        ape = ae / np.maximum(np.abs(y_true), epsilon)
        # When true values are very small, we take MAE
        small_y_mask = y_true < epsilon
        ape = np.where(small_y_mask, ae, ape)
    else:
        raise ValueError(f'Undefined zeros_strategy {zeros_strategy}')

    return np.mean(ape)


def plot_results(plotting, train, test, X_dummies_train, target, gs, model, model_name, i, mae_test):
    if plotting:
        print('Plot')
        # Features
        # TODO SHAP
        no_features_imp = [
            'SVR', 'ExpSmoothingRegressor', 'NaiveRegressor',
            'MLPRegressor', 'ANNRegressor', 'LSTMRegressor', 'TCNRegressor'
        ]
        if model_name not in no_features_imp:
            if model_name == "LinearRegression":
                impact = model.named_steps['regressor'].coef_
            else:
                impact = model.named_steps['regressor'].feature_importances_

            # not_dummies = [i for el in model.named_steps['preprocessor'].transformers_
            # if hasattr(el[1], 'get_feature_names') for i in el[1].get_feature_names()]
            features_df = pd.DataFrame({
                'impact': impact,
                'feature': list(X_dummies_train.columns),
            })

            fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(20, 12))
            features_df.sort_values('impact', ascending=True).plot.barh(x='feature', y='impact', ax=ax)
            fig.tight_layout(pad=4.0)
            fig.savefig(f'{model_name}_features.png')


        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(20, 12))

        # CV errors
        pred_cv_features = [f'pred_cv_{j}' for j in range(i)]
        mae_cv = abs(gs.best_score_)
        subset_cv = train.loc[train['pred_cv'].notnull(), [target] + pred_cv_features]
        subset_cv.plot(y=[target] + pred_cv_features,
                       title='CV MAPE: {0:.0%}'.format(mae_cv), ax=axes[0][0])

        # Train errors:
        train['pred'] = model.predict(X_dummies_train)
        mae = mean_absolute_percentage_error(train[target], train['pred'])
        subset_train = train[[target, 'pred']]
        subset_train.plot(y=[target, 'pred'],
                          title='Train MAPE: {0:.0%}'.format(mae), ax=axes[0][1])

        # Test errors zoomed
        subset_test = test[[target, 'pred', 'pi_lower', 'pi_upper']]
        subset_test.plot(y=[target, 'pred', 'pi_lower', 'pi_upper'],
                         title='Testing MAPE: {0:.0%}'.format(mae_test), ax=axes[1][0])

        # Test errors
        rejoined = subset_train.rename(columns={target: 'train'})[['train']].join(
            subset_test.rename(columns={target: 'test'})[['test', 'pred', 'pi_lower', 'pi_upper']], how='outer')
        rejoined.plot(y=['train', 'pred', 'test'],
                      title='Testing MAPE: {0:.0%}'.format(mae_test), ax=axes[1][1])

        for ax_sub in axes:
            for ax in ax_sub:
                ax.spines['top'].set_visible(False)
                ax.spines['right'].set_visible(False)

        fig.tight_layout(pad=4.0)
        fig.savefig(f'{model_name}.png')

        # fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(20, 12))
        # rejoined.plot(y=['train', 'pred', 'test', 'pi_lower', 'pi_upper'],
        #               title='Testing MAPE: {0:.0%}'.format(mae_test), ax=ax)
        # ax.spines['top'].set_visible(False)
        # ax.spines['right'].set_visible(False)
        # fig.tight_layout(pad=4.0)
        # fig.savefig(f'{model_name}_prediction.png')

        return rejoined


# noinspection PyDefaultArgument
def check_model(regressor, params, dataset,
                target='new_cases', features=['date'], categorical_features=[], user_transformers=[],
                h=14, n_splits=5, gap=14,
                plotting=True
                ):
    """
    Check model
    :param gap:
    :param regressor:
    :param params:
    :param dataset:
    :param target:
    :param features:
    :param categorical_features:
    :param user_transformers:
    :param h:
    :param n_splits:
    :param plotting:
    :return:
    """
    # TODO info about dropped
    y, X, X_dummies = separate_data(dataset.dropna(), target, features, categorical_features)
    train, test = split(dataset.dropna(), h)
    X_dummies_train, X_dummies_test = split(X_dummies, h)
    y_train, y_test = split(y, h)

    cv = custom_split(train, h, n_splits=n_splits, gap=gap)

    # https://newbedev.com/sklearn-pipeline-get-feature-names-after-onehotencode-in-columntransformer
    # https://stackoverflow.com/questions/57528350/can-you-consistently-keep-track-of-column-labels-using-sklearns-transformer-api/57534118#57534118
    # https://scikit-learn.org/stable/tutorial/statistical_inference/putting_together.html
    # categorical_transformer = OneHotEncoder(handle_unknown='ignore')
    preprocessor = ColumnTransformer(
        transformers=[('num', FunctionTransformer(), list(X_dummies.columns))] +
                     [(ut[0], ut[1], list(X_dummies.columns)) for ut in user_transformers]
    )
    pipeline = Pipeline(steps=[('preprocessor', preprocessor), ('regressor', regressor)])
    # rename params:
    params = {f'regressor__{k}': v for k, v in params.items()}

    # search parameters space
    # TODO change to log
    print('Grid search')
    gs = GridSearchCV(
        estimator=pipeline,
        param_grid=params,
        scoring=make_scorer(mean_absolute_percentage_error, greater_is_better=False),
        cv=cv[0]
    )
    # print(X_dummies_train.head())
    gs.fit(X_dummies_train, y_train)
    model = gs.best_estimator_

    # check cv results
    i = 0
    train['pred_cv'] = None
    for train_split, test_split in cv[1]:
        train[f'pred_cv_{i}'] = None
        # TODO prevent retraining
        model.fit(X_dummies_train.loc[train_split, :], y_train.loc[train_split])
        train.loc[test_split, f'pred_cv_{i}'] = model.predict(X_dummies_train.loc[test_split, :])
        train.loc[test_split, 'pred_cv'] = model.predict(X_dummies_train.loc[test_split, :])
        i += 1

    # fit model to train
    print('Fitting to train')
    model.fit(X_dummies_train, y_train)

    # check performance on test dataset
    test['pred'] = model.predict(X_dummies_test)
    if hasattr(model.named_steps["regressor"], 'predictions'):
        test['pi_lower'] = model.named_steps["regressor"].predictions['pi_lower']
        test['pi_upper'] = model.named_steps["regressor"].predictions['pi_upper']
    else:
        test['pi_lower'] = None
        test['pi_upper'] = None

    mape_test = mean_absolute_percentage_error(test[target], test['pred'])
    mae_test = mean_absolute_error(test[target], test['pred'])
    rmse_test = math.sqrt(mean_squared_error(test[target], test['pred']))

    model_name = type(model.named_steps["regressor"]).__name__
    rejoined = plot_results(plotting, train, test, X_dummies_train, target, gs, model, model_name, i, mape_test)

    # TODO remove regressor_ from best_params

    return {
        'model_name': type(pipeline.named_steps["regressor"]).__name__,
        'model': model,
        'mape_cv': abs(gs.best_score_),
        'std_cv': pd.DataFrame(gs.cv_results_).sort_values('rank_test_score')['std_test_score'].values[0],
        'performance_test': {'MAE': mae_test, 'MAPE': mape_test, 'RMSE': rmse_test},
        'best_params': gs.best_params_,
        'cv_results': pd.DataFrame(gs.cv_results_),
        'rejoined': rejoined,
    }


